# mcdu control file. This one does all the graphical work.
# $Id: mcdu.tcl,v 1.42 2008/03/09 14:07:10 hoppie Exp $

# Copyright (C) 2010 Jeroen Hoppenbrouwers <hoppie@hoppie.nl>
# This program is free software; more information in the LICENSE file.


namespace eval mcdu {
  # Module-wide. NOTE: some of these likely can be moved to private.
  variable DISPLAY_MODE ""         ; # Display mode from /etc/*.cfg.
  variable FONT_SIZE 10            ; # Font size, calculated from dsply mode.
  variable FONT_SMALL              ; # Calculated string: small font ID.
  variable FONT_LARGE              ; # Calculated string: large font ID.
  variable ROWS 14                 ; # Number of display rows, cached.
  variable COLS 24                 ; # Number of display columns, cached.
  variable META 0                  ; # Number of META rows under display rows.
  variable DEFAULTCOLOUR {}        ; # Contains default colour in {r g b}.
  variable COLOURS                 ; # array(row,col) of base colour values.
  variable MESSAGES {}             ; # List with messages.
  variable SCRATCHPAD ""           ; # Current scratch path contents.
  variable COUNTER [rnd 100]       ; # Key counter for Broker trickery.
  variable MODULES {}              ; # List with currently available modules.
  variable LAYOUT                  ; # (screenpos)="module" with screenpos
  array set LAYOUT {}              ; # 0..10, LSK1L to 5R.
  variable FUNCTIONKEYS            ; # Array (functionKey)=moduleName.
  variable RESET_ARMED 0           ; # 1 when reset is armed (config page).
  variable KEYBOARD ""             ; # Keyboard code of the mapping table.
  variable QUIT_ARMED 0            ; # 1 when quit is armed (config page).
  variable SHUTDOWN_ARMED 0        ; # 1 when shutdown is armed (config page).
  variable PASSWORD_OK 0           ; # 1 when password ok (config page).
  variable DRAG {0 0}              ; # X and y where window mouse drag started.
  variable DRAGICON {0 0}          ; # X and y where icon mouse drag started.
  variable DRAGNOTYET 0            ; # Delay timer for the icon mouse drag.
  variable ICONIC 0                ; # 0 when normal, 1 when iconic.
  variable POWER -1                ; # 0 when powerless, 1 when powered.
                                   ; # -1 means ignore at this moment.
  variable PASSTHROUGH             ; # Array (modName) = 1 for passthrough mode,
  array set PASSTHROUGH {}         ; # 0 for normal mode.

  variable XLAT

  # Private (static).
  variable _MODE ""               ; # Current mode.
  variable _KEYINT 0              ; # Key press handling in progress.
  variable _ANNUNCIATORS 0        ; # Timing semaphore.
  variable _ANNUNCIATORS_MAGENTA 0; # Timing semaphore.
  variable _TEST 0                ; # Current test display mode.
}


proc mcdu::done {} {
  # End the MCDU module.

  # Extinghuish the annunciators and pause a bit.
  ann all 0
  message add "mcdu quit"
  update
  sleep 1
  powerEvent 1

  # Send a keystroke with the RESET mode. The keystroke itself is a dummy.
  mode reset
  sendKey lsk1l
}; # done


proc mcdu::init {} {
  # Set up all internal data structures and build the main GUI.
  global IMGDIR
  global tcl_platform
  variable FONT_SIZE
  variable FONT_SMALL
  variable FONT_LARGE
  variable ROWS
  variable COLS
  variable DEFAULTCOLOUR
  variable COLOURS

  # Throw away possibly existing widgets. TODO: consider a subframe for all
  # button and annunciator widgets, instead of root path names.
  destroy .c
  destroy .lsk1l
  destroy .lsk2l
  destroy .lsk3l
  destroy .lsk4l
  destroy .lsk5l
  destroy .lsk6l
  destroy .lsk7l
  destroy .lsk8l
  destroy .lsk9l
  destroy .lsk10l
  destroy .lsk11l
  destroy .lsk1r
  destroy .lsk2r
  destroy .lsk3r
  destroy .lsk4r
  destroy .lsk5r
  destroy .lsk6r
  destroy .lsk7r
  destroy .lsk8r
  destroy .lsk9r
  destroy .lsk10r
  destroy .lsk11r
  destroy .menu
  destroy .nextpage
  destroy .prevpage
  destroy .clr
  destroy .del
  destroy .icon

  # Throw away any possibly existing images, to avoid eating up memory.
  foreach image [image names] {
    image delete $image
  }

  # Fetch the display mode settings.
  array set mode [readMode [prefs::get display.mode]]

  # Get display rows and columns.
  set ROWS [lindex $mode(layout) 0]
  set COLS [lindex $mode(layout) 1]

  # Get Meta Block number of rows (below the lowest display row).
  set META [lindex $mode(layout) 2]
  if {$META==""} {set META 0}

  # Initialise the COLOURS matrix with the MCDU default display colour from
  # the config file. We only need a full matrix for the setDisplayBrightness
  # function; all others do not look here at all, but use the pen colour.
  set DEFAULTCOLOUR $mode(colour)
  for {set r 0} {$r<=$ROWS} {incr r} {
    for {set c 0} {$c<=$COLS} {incr c} {
      set COLOURS($r,$c) $DEFAULTCOLOUR
    }
  }

  # Get character cell column width and row height from display rectangle.
  set displayX1 [lindex $mode(display) 0]
  set displayY1 [lindex $mode(display) 1]
  set displayX2 [lindex $mode(display) 2]
  set displayY2 [lindex $mode(display) 3]
  set displayWidth  [expr $displayX2-$displayX1]
  set displayHeight [expr $displayY2-$displayY1]
  set cw [expr $displayWidth.0/$COLS]
  set rh [expr $displayHeight.0/($ROWS+$META)]
  # Calculate the font size in whole pixels for future use by mcdu::display.
  set FONT_SIZE [expr round(($displayWidth/16)*-1.12)]

  # Calculate the font strings for small and large fonts.
  set FONT_SMALL "$mode(font.small.family)\
                 [expr round($FONT_SIZE*$mode(font.small.factor))]\
                 $mode(font.small.style)"
  set FONT_LARGE "$mode(font.large.family)\
                 [expr round($FONT_SIZE*$mode(font.large.factor))]\
                 $mode(font.large.style)"

  # Fundamental selection: full screen or photo-realistic.
  if {[info exists mode(image)] && \
      [file exists "$IMGDIR/$mode(image).gif"]} {

    # Mode contains an image clause and the image is available.
    # Bitmap required, and default normal window manager decorations.
    if {$mode(titlebar.hide)=="default"} {
      set mode(titlebar.hide) 0
    }
    # State switches to avoid artifacts on Linux.
    wm state . withdraw
    wm overrideredirect . $mode(titlebar.hide)
    wm state . normal

    # We call the loaded image "background".
    image create photo background -file "$IMGDIR/$mode(image).gif" -format gif

    # The canvas has the exact same size as the bitmap it must contain.
    set canvasWidth  [image width background]
    set canvasHeight [image height background]
    canvas .c -width $canvasWidth -height $canvasHeight -background black
    .c create image 0 0 -image background -anchor nw

    # Offsets for visual centering on bitmap display.
    set dx [expr round($displayX1+$FONT_SIZE*-0.25)]
    set dy $displayY1

    # Blank out the display part of the image (needs to be overwritten).
    .c create rectangle $displayX1 $displayY1 \
                        $displayX2 $displayY2 -fill black

    # All annunciators come as overlay bitmaps, too. Load them up.
    foreach ann {exec dspy fail msg ofst} {
      if {[info exists mode(ann_$ann)]} {
        set bitmap [lindex $mode(ann_$ann) 0]
        if {[file exists "$IMGDIR/$bitmap.gif"]} {
          image create photo ann_$ann -file "$IMGDIR/$bitmap.gif" -format gif
        }
      }
    }; # foreach annunciator
    
  } else {

    # Mode does not contain an image clause.
    # No bitmap required, just a display with no window manager decorations.
    if {$mode(titlebar.hide)=="default"} {
      set mode(titlebar.hide) 1
    }
    # State switches to avoid artifacts on Linux.
    wm state . withdraw
    wm overrideredirect . $mode(titlebar.hide)
    wm state . normal

    # Canvas is exact display size.
    canvas .c -width  $displayWidth \
              -height $displayHeight \
              -background black
    # Offsets for visual centering on hardware display.
    set dx [expr round($FONT_SIZE*-0.375)]
    set dy 0
  }; # Selection between photo-realistic and full screen display.

  # Show the canvas in the GUI.
  pack .c -fill both -expand 1

  # Create a character matrix on the canvas for the display. The colours are
  # added by the subsequent setDisplayBrightness call, which needs a
  # complete COLOURS array (which has been set up above).
  # rows/cols = number of rows/columns to create.
  # cw/rh = column width/row height in pixels.
  # dx/dy = offset from the top-left corner of the canvas in pixels.
  for {set row 0} {$row<$ROWS} {incr row} {
    for {set col 0} {$col<$COLS} {incr col} {
      .c create text \
                [expr ($col)*$cw+$dx] \
                [expr ($row+1)*$rh+$dy] \
                -text " " -tags "$row,$col" -anchor s
    }
  }

  # Use cell (-1,-1) to store the meta block colour, which is default.
  setCellColour -1 -1
  # Meta Block required?
  if {$META>0} {
    # Yes, create Meta Block on the display canvas below the character
    # matrix.
    set meta(x)      $dx
    set meta(y)      [expr $row*$rh+$dy]
    set meta(width)  [expr ($displayWidth-$dx)/6]
    set meta(height) [expr $rh*2]
    set meta(font)   "Arial [expr $FONT_SIZE/2]"

    metaBlock 0 "MCDU\nMENU" [array get meta]
    metaBlock 1 "PREV\nPAGE" [array get meta]
    metaBlock 2 "NEXT\nPAGE" [array get meta]
    metaBlock 3 "KEY\nBRD"   [array get meta]
    metaBlock 4 ""           [array get meta]
    metaBlock 5 "CLEAR"      [array get meta]

    metaAnn 4 1 "FAIL" [array get meta]
    metaAnn 4 2 "MSG"  [array get meta]
  }

  setDisplayBrightness

  # Array of hot spots to catch mouse clicks on key areas. They are created
  # last, so they are on top of the Z-order.
  # This procedure only defines hot spots the size of a point. The general
  # mouse click handler uses a "closest" function to determine whether there
  # was a click "near" one of these hot spots.
  foreach keysym {lsk1l lsk2l lsk3l lsk4l lsk5l lsk6l lsk7l lsk8l lsk9l
                  lsk10l lsk11l
                  lsk1r lsk2r lsk3r lsk4r lsk5r lsk6r lsk7r lsk8r lsk9r
                  lsk10r lsk11r
                  initref rte deparr atc vnav brt- brt+
                  fix legs hold fmccomm prog exec
                  menu navrad
                  prevpage nextpage
                  a b c d e f g h i j k l m n o p q r s t u v w x y z
                  sp del / clr
                  1 2 3 4 5 6 7 8 9 0 . +/-
                  kbd} {
    if {[info exists mode(key_$keysym)]} {
      # x and y are the center coordinates of the master hot spot.
      set x [lindex $mode(key_$keysym) 0]
      set y [lindex $mode(key_$keysym) 1]
      # repeat, if available, is the number of slave hot spots to draw from the
      # master hot spot to the right, or to the left when negative.
      set repeat [lindex $mode(key_$keysym) 2]
      # Create master hot spot (always).
      .c create bitmap $x $y -tag "KEY_$keysym"
      if {$repeat!=""} {
        # Create slave hot spots to the left or the right of the master.
        # Remember that FONT_SIZE is always negative (pixels, not pica
        # points). And also that the "find closest" halo is circular.
        if {$repeat<0} {
          set repeat [expr -$repeat]
          for {set i 1} {$i<=$repeat} {incr i} {
            .c create bitmap [expr $x+($i*$FONT_SIZE*1.0)] $y -tag "KEY_$keysym"
          }
        } else {
          for {set i 1} {$i<=$repeat} {incr i} {
            .c create bitmap [expr $x-($i*$FONT_SIZE*1.0)] $y -tag "KEY_$keysym"
          }
        }
      }; # if there is a string of slave hot spots
    }; # if there is a master hot spot
  }; # foreach keysym

  # Bind a handler to all key presses. Use a "break" to prevent Windows from
  # co-processing F10 and other hot keys that may play tricks.
  bind . <KeyPress> {mcdu::keyBoard %K %k; break}
  bind . <Control-KeyPress> {mcdu::keyBoard %K %k Control; break}

  # Bind a handler to all mouse clicks.
  bind . <Button> {mcdu::mouse %x %y %b}

  # Bind a handler to all mouse motion.
  bind . <Motion> mcdu::motion

  # Bind a handler to MB3 (right button) dragging.
  bind . <Button3-Motion> {mcdu::drag %x %y}

  # Install the correct keyboard map.
  keyboard [prefs::get keyboard]

  # Some "unneccesary" things that Windows asks for.
  wm resizable . 0 0
  update idletasks
  raise .
  focus -force .

  # Hide mouse pointer if required by settings.
  mouseHide

  if {$mode(floating.icon)} {
    # Create an always-on-top icon window.
    toplevel .icon
    wm overrideredirect .icon 1
    # The PID is attached to allow multiple MCDU instances on the same
    # window system.
    wm title .icon "MCDU-ICON-[pid]"
    pack [frame .icon.f]
    pack [label .icon.f.l -text "> CDU" -foreground green -background black]\
      -ipadx 10 -ipady 5
    bind .icon <1> {mcdu::mouseIcon %x %y}
    bind .icon <Double-1> {mcdu::mouseIcon %x %y double}
    bind .icon <Button1-Motion> {mcdu::dragIcon %x %y}

    # Position the icon on the screen at its last coordinates, or 
    # leave its position to the window manager if they are unknown.
    set geom [prefs::get icon.position]
    if {$geom!=""} {
      wm geometry .icon $geom
    }

    if {$tcl_platform(platform)=="unix"} {
      # On Linux (at least Ubuntu 10.10), we need to force override redirect
      # for the main toplevel (MCDU) window here, else it will get higher
      # on-top priority as the icon. Weird!
      wm overrideredirect . 0
    }
    wm attributes .icon -topmost 1

    # Create the MB3 popup menu for the floating icon...
    set m [menu .icon.menu -tearoff 0]
    $m add command -label "Exit MCDU" -underline 1 -command gui::done
    # ... and bind it to MB3 on the icon.
    bind .icon <3> {tk_popup .icon.menu %X %Y}

  }; # if icon window required

  # Start reacting on power events after a while only.
  after 5000 {set mcdu::POWER 1}
}; # init


proc mcdu::metaBlock {pos text metaList} {
  # Paint a meta block.
  # <in> pos = Meta 'column' position, 0..n.
  # <in> text = Text to paint, may have \n inside or be empty.
  # <in> meta = List from array with basic positioning data.
  array set meta $metaList
  set x1 [expr $meta(x)+$pos*$meta(width)+2]
  set y1 [expr $meta(y)+4]
  set x2 [expr $meta(x)+($pos+1)*$meta(width)-4]
  set y2 [expr $meta(y)+$meta(height)-2]
  .c create rectangle $x1 $y1 $x2 $y2 -fill [getCellColour -1 -1] -tag meta
  .c create text \
            [expr $x1+$meta(width)/2] \
            [expr $y1+$meta(height)/2] \
            -text $text -fill black -font $meta(font)
}; # metaBlock


proc mcdu::metaAnn {pos stack text metaList} {
  # Paint a meta annunciator.
  # <in> pos = Meta 'column' position, 0..n.
  # <in> stack = Meta 'row' position, 1..2.
  # <in> text = Annunciator text.
  # <in> meta = List from array with basic positioning data.
  array set meta $metaList
  set x [expr $meta(x)+$pos*$meta(width)+2]
  set y [expr $meta(y)+4]

  set tag [string tolower $text]
  .c create text \
            [expr $x+$meta(width)/2] \
            [expr $y+$stack*$meta(height)/3] \
            -text $text -fill [getCellColour -1 -1] \
            -font $meta(font) -tag "meta ann_meta_$tag"
}; # metaAnn


proc mcdu::virtualKeyboard {} {
  # Pop up a virtual keyboard for occasional alphanum entries. If the
  # keyboard already exists, remove popup.
  variable FONT_SIZE

  # Hide the mouse pointer when needed. A small delay is used to suppress
  # finger motion after tap on a touch screen.
  after 500 {mcdu::mouseHide}

  if {[winfo exists .vkbd]} {
    .c delete vkbd
    destroy .vkbd
    return
  }

  # Virtual keyboard: lots of buttons in frames. Alpha block and num block,
  # separated by a bit of space. Both blocks with four rows of 8 and 3 cols.
  # Note that the top frame .vkbd will be displayed as window in the .c
  # canvas, not packed.
  set k [frame .vkbd -relief raised -border 3]
  set font "Arial [expr $FONT_SIZE/2]"

  # Alpha block.
  set a [frame $k.a]
  foreach {frame charList} {1 {A B C D E F G H}
                            2 {I J K L M N O P}
                            3 {Q R S T U V W X}
                            4 {Y Z sp del / clr}} {
    pack [frame $a.$frame] -side top -pady 5
      foreach char $charList {
        pack [button $a.$frame.[string tolower $char] \
              -text $char -width 2 -font $font \
              -command "mcdu::keyPressed [string tolower $char]"] \
          -side left -padx 5
      }
  }
  # The last two buttons are spares. May be integrated in .4 block when
  # used.
  foreach char {1 2} {
    pack [button $a.4.$char -width 2 -font $font] -side left -padx 5
  }
  
  # Numeric block.
  set n [frame $k.n]
  foreach {frame charList} {1 {1 2 3}
                            2 {4 5 6}
                            3 {7 8 9}} {
    pack [frame $n.$frame] -side top -pady 5
    foreach char $charList {
      pack [button $n.$frame.$char -text $char -width 2 -font $font \
            -command "mcdu::keyPressed $char"] \
        -side left -padx 5
    }
  }
  # . and +/- cannot be used as widget names.
  pack [frame $n.4] -side top -pady 5
  pack [button $n.4.dot -text . -width 2 -font $font \
        -command "mcdu::keyPressed ."] -side left -padx 5
  pack [button $n.4.0 -text 0 -width 2 -font $font \
        -command "mcdu::keyPressed 0"] -side left -padx 5
  pack [button $n.4.pm -text "+/-" -width 2 -font $font \
        -command "mcdu::keyPressed +/-"] -side left -padx 5

  # Alpha and num block side by side with spacing.
  pack $a $n -padx 5 -pady 5 -side left

  # Fetch the display geometry.
  array set mode [readMode [prefs::get display.mode]]
  set displayX1 [lindex $mode(display) 0]
  set displayY1 [lindex $mode(display) 1]
  set displayX2 [lindex $mode(display) 2]
  set displayY2 [lindex $mode(display) 3]
  set displayWidth  [expr $displayX2-$displayX1]
  set displayHeight [expr $displayY2-$displayY1]

  # Virtual keyboard rectangle is positioned horizontally centered and one
  # line below the top of the display. Vertical size depends only on the
  # FONT_SIZE.
  set vkbdX [expr $displayX1+($displayWidth/2)]
  set vkbdY [expr $displayY1-$FONT_SIZE*4]
  .c create window $vkbdX $vkbdY -window $k -tag vkbd
}; # virtualKeyboard


proc mcdu::setDisplayBrightness {} {
  # Reads the display.brightness setting and then resets the correct RGB
  # colours for all individual display cells. The COLOURS matrix is used to
  # find the base colour for each individual cell. It is set up with the
  # config file default colour at MCDU init time, and modified per cell with
  # the MCDU colour API calls while drawing characters.
  variable ROWS
  variable COLS
  variable COLOURS

  # Run over all rows and columns and change the colour.
  for {set row 0} {$row<$ROWS} {incr row} {
    for {set col 0} {$col<$COLS} {incr col} {
      .c itemconfigure "$row,$col" -fill [getCellColour $row $col]
    }
  }

  # Also update all Meta Block items (if any).
  .c itemconfigure "meta" -fill [getCellColour -1 -1]
  update idletasks
}; # setDisplayBrightness


proc mcdu::setCellColour {row col {colour ""}} {
  # Sets the base colour for the given cell. Via getCellColour the actual
  # (adjusted for brightness) cell colour can then be retrieved.
  # <in> row = Row to set the colour for, 0..n.
  # <in> col = Column to set the colour for, 0..n.
  # <in> colour = {red green blue} base colour for this cell. If empty,
  #               uses current DEFAULTCOLOUR.
  variable COLOURS
  variable DEFAULTCOLOUR
  if {$colour==""} {set colour $DEFAULTCOLOUR}
  set COLOURS($row,$col) $colour
}; # setCellColour


proc mcdu::getCellColour {row col} {
  # Calculate the actual cell colour, i.e., including the brightness
  # adjustment.
  # <in> row = Cell row, 0..n.
  # <in> col = Cell column, 0..n.
  # <out> The colour in '#rrggbb' format, ready for 'canvas itemconfigure'.
  variable COLOURS

  set colour $COLOURS($row,$col)
  set maxRed   [lindex $colour 0]
  set maxGreen [lindex $colour 1]
  set maxBlue  [lindex $colour 2]

  # Brightness varies between 0 and 10, default 7.
  set brightness [prefs::get display.brightness]
  if {$brightness==""} {set brightness 7}

  # Use a nonlinear scale to adjust for human eye behaviour.
  set brightness \
      [expr round(($brightness*$brightness*0.8)+$brightness*1.5+5.0)]

  # Calculate the actual RGB colours of the cell using the brightness.
  set red   [expr $maxRed*$brightness/100]
  set green [expr $maxGreen*$brightness/100]
  set blue  [expr $maxBlue*$brightness/100]
  set rgb [format {#%02x%02x%02x} $red $green $blue]

  return $rgb
}; # getCellColour


proc mcdu::readMode {mode} {
  # Read in the requested mode file, and return the content as a
  # Tcl list in "array set" syntax.
  # <in> mode = The requested mode.
  # <out> The mode array in "array set" syntax, (key)=value.
  global ETCDIR
  
  # Check whether the mode file is available.
  if {![file exists "$ETCDIR/$mode.cfg"]} {
    # It is not available; get another one. If that fails, panic.
    set mode [config_nextMode ""]
    if {$mode==""} {die "No mode files available; check etc/*.cfg"}
  }

  # Read the mode file. Provide for some default values.
  set f [open "$ETCDIR/$mode.cfg" "r"]
  array set settings {
    display {0 0 640 480}
    colour {115 255 115}
    layout {14 24}
    mouse.autohide 0
    titlebar.hide default
    floating.icon 0
    menu.skip 0
    splash.skip 0
    config.prompt ""
    config.password ""
    font.small.family {{Aerowinx FMC Small}}
    font.small.style ""
    font.small.factor 1.0
    font.large.family {{Aerowinx FMC Large}}
    font.large.style ""
    font.large.factor 0.95
  }
  while {![eof $f]} {
    set line [string trim [gets $f]]
    if {$line==""} continue;
    if {[string index $line 0]=="#"} continue;
    set key [lindex $line 0]
    set value [lrange $line 1 end]
    set settings($key) $value
  }
  close $f
  return [array get settings]
}; # readMode


proc mcdu::keyboard {type} {
  # Install the keyboard map.
  # <in> type = The required map name. If empty, uses "PS1". Currently
  #             supported are "off", "magenta", and "PS1".
  variable XLAT

  catch {unset XLAT}

  switch $type {
    off {
      # Mode "off" completely disables the keyboard. Handy for some specific
      # applications. Switch on again by selecting one of the other modes.
      array set XLAT {}
    }
    magenta {
      # Precision Flight Controls MCDU key stroke table.
      array set XLAT {
	8 clr
	13 exec
	32 sp
	33 prevpage
	34 nextpage
        111 brt+
        40 brt-
	46 del
	48 0
	49 1
	50 2
	51 3
	52 4
	53 5
	54 6
	55 7
	56 8
	57 9
	65 a
	66 b
	67 c
	68 d
	69 e
	70 f
	71 g
	72 h
	73 i
	74 j
	75 k
	76 l
	77 m
	78 n
	79 o
	80 p
	81 q
	82 r
	83 s
	84 t
	85 u
	86 v
	87 w
	88 x
	89 y
	90 z
	112 lsk1l
	113 lsk2l
	114 lsk3l
	115 lsk4l
	116 lsk5l
	117 lsk6l
	118 lsk1r
	119 lsk2r
	120 lsk3r
	121 lsk4r
	122 lsk5r
	123 lsk6r
	187 +/-
	190 .
	191 /
	Control-112 initref
	Control-113 rte
	Control-114 deparr
        Control-115 atc
	Control-116 vnav
	Control-117 fix
	Control-118 legs
	Control-119 hold
        Control-120 fmccomm
	Control-121 prog
	Control-122 menu
	Control-123 navrad
      }
    }
    linux {
      # Key strokes on a typical Linux distribution.
      array set XLAT {
	22 clr
	118 exec
	65 sp
	112 prevpage
	117 nextpage
        111 brt+
        116 brt-
	115 navrad
	110 menu
	119 del
	90 0
	87 1
	88 2
	89 3
	83 4
	84 5
	85 6
	79 7
	80 8
	81 9
	38 a
	56 b
	54 c
	40 d
	26 e
	41 f
	42 g
	43 h
	31 i
	44 j
	45 k
	46 l
	58 m
	57 n
	32 o
	33 p
	24 q
	27 r
	39 s
	28 t
	30 u
	55 v
	25 w
	53 x
	29 y
	52 z
	73 initref
	74 rte
	75 deparr
	76 vnav
        95 atc
	16 fix
	17 legs
	18 hold
	19 prog
        20 fmccomm
	67 lsk1l
	68 lsk2l
	69 lsk3l
	70 lsk4l
	71 lsk5l
	72 lsk6l
	10 lsk1r
	11 lsk2r
	12 lsk3r
	13 lsk4r
	14 lsk5r
	15 lsk6r
	82 +/-
	91 .
	106 /
      }
    }
    default {
      # PS1 key stroke table.
      array set XLAT {
	8 clr
	32 sp
	33 prevpage
	34 nextpage
	35 navrad
	36 menu
        38 brt+
        40 brt-
	45 exec
	46 del
	48 prog
	49 lsk1l
	50 lsk2l
	51 lsk3l
	52 lsk4l
	53 lsk5l
	54 lsk6l
	55 fix
        56 legs
        57 hold
	65 a
	66 b
	67 c
	68 d
	69 e
	70 f
	71 g
	72 h
	73 i
	74 j
	75 k
	76 l
	77 m
	78 n
	79 o
	80 p
	81 q
	82 r
	83 s
	84 t
	85 u
	86 v
	87 w
	88 x
	89 y
	90 z
	96 0
	97 1
	98 2
	99 3
	100 4
	101 5
	102 6
	103 7
	104 8
	105 9
	109 +/-
	110 .
	111 /
	112 lsk1r
	113 lsk2r
	114 lsk3r
	115 lsk4r
	116 lsk5r
	117 lsk6r
	118 initref
	119 rte
	120 deparr
	121 vnav
      }
    }
  }; # switch type
}; # keyboard


proc mcdu::display {row col text {size ""}} {
  # Fully equivalent to mcdu::displayc but without the colour parameter.
  # The colour can be set with mcdu::colour if required.
  displayc $row $col "" $text $size
}; # display


proc mcdu::displayc {row col colour text {size ""}} {
  # Display a text starting at the row and column in font size, using the
  # default drawing colour.
  # The given text does not wrap to the next row when exceeding display
  # width. Typical displays have 14 rows of 24 characters, but may have more.
  # <in> row = 0..13 (top = 0, bottom most = -1).
  # <in> col = 0..23 left aligned, -23..-1 right aligned. Or "l" for left
  #            aligned, "c" for centered, and "r" for right aligned.
  # <in> colour = The display colour, one of the ARINC 739 colours or "" for
  #               no change from the last used colour. "default" means the
  #               display mode config default colour (greenish, usually).
  #               One of: black,cyan,red,yellow,green,magenta,amber,white.
  #               Default colour changes to this one after the call.
  # <in> text = The character(s) to display, possibly between "".
  #             Special characters:
  #               ! for boxes
  #               ' for degree symbols
  #             Most non-alphanum characters are not displayed.
  # <in> size = Optionally, font size, "small" or "large". If left out,
  #             the default font size for the line is selected.
  variable FONT_SIZE
  variable FONT_SMALL
  variable FONT_LARGE
  variable ROWS
  variable COLS
  variable PASSTHROUGH

  # Special column alignment?
  switch -- [string tolower $col] {
    "l" {set col 0}
    "c" {set col [expr int(12-([string length $text]/2.0))]}
    "r" {set col -[string length $text]}
  }

  # Sanity check on row/col figures to prevent runtime errors.
  if {[regexp {^-?[0123456789]+$} $row]==0} {
    message add "Illegal row $row"
    return
  }
  if {[regexp {^-?[0123456789]+$} $col]==0} {
    message add "Illegal column $col"
    return
  }

  # Wrap around if negative values are given.
  if {$row<0} {incr row $ROWS}
  if {$col<0} {incr col $COLS}

  # Change the colour if not left out.
  if {$colour!=""} {
    colour $colour
  }

  # Default font switches. Even lines support both font sizes.
  if {$size==""} {
    if {$row==[expr $ROWS-1]} {
      # Scratch pad.
      set size "large"
    } elseif {[expr $row%2]} {
      # Label line.
      set size "small"
    } else {
      # Data line.
      set size "large"
    }
  }

  if {$size=="small"} {
    set font $FONT_SMALL
  } else {
    set font $FONT_LARGE
  }

  # Paint all text on the line. Capitals only.
  set text [string toupper $text]

  set convert 1
  if {[info exists PASSTHROUGH([mode])]} {
    set convert [expr !$PASSTHROUGH([mode])]
  }

  for {set c 0} {$c<[string length $text]} {incr c} {
    set ch [string index $text $c]
    if {$convert} {
      # In normal (non-passthrough) mode, convert many non-ASCII characters.
      if {[regexp {[^0-9A-Z!%()+\-/'><.#*^?:;~]} $ch]} {set ch " "}
      switch -- $ch {
        "'" {set ch "g"}
        "^" {set ch "u"}
        "?" {set ch "d"}
        ":" {set ch "l"}
        ";" {set ch "r"}
        "~" {set ch "t"}
      }
    }
    setCellColour $row [expr $col+$c]
    .c itemconfigure "$row,[expr $col+$c]" \
       -text $ch \
       -font $font \
       -fill [getCellColour $row [expr $col+$c]]
  }

  # Tell Engravity driver that this row changed.
  engr::display $row
}; # displayc


proc mcdu::colour {{colour ""}} {
  # Set the paint colour for all mcdu::display calls from now on.
  # Overrides the colour setting of the mode config file; reverts to default
  # (mode config file) colour if called with no parameters.
  # <in> colour = Optionally, new paint colour.
  variable DEFAULTCOLOUR

  # These colours come straight from ARINC 739A-1, 'the' MCDU spec.
  switch $colour {
    black   {set DEFAULTCOLOUR {  0   0   0}}
    cyan    {set DEFAULTCOLOUR {  0 255 255}}
    red     {set DEFAULTCOLOUR {255   0   0}}
    yellow  {set DEFAULTCOLOUR {255 255   0}}
    green   {set DEFAULTCOLOUR {  0 255   0}}
    magenta {set DEFAULTCOLOUR {255   0 255}}
    amber   {set DEFAULTCOLOUR {255 128   0}}
    white   {set DEFAULTCOLOUR {255 255 255}}
    default {
      array set mode [readMode [prefs::get display.mode]]
      set DEFAULTCOLOUR $mode(colour)
    }
  }
}; # colour


proc mcdu::readRow {row} {
  # Returns the current character content of the requested row, as a
  # character string with alternating character and font size:
  # <SLELLLELCLTL etc. (L and S for font size).
  # TODO: the font size is currently fixed at "L" as only the Engravity MCDU
  # unit driver uses the readRow() procedure.
  # See also: getRow() and getLabel().
  # <in> row = The row to read, typically 0..13.
  # <out> The read row, not trimmed, so of fixed length.
  variable COLS
  set rowString ""
  for {set c 0} {$c<$COLS} {incr c} {
    append rowString "[.c itemcget "$row,$c" -text]L"
  }
  return $rowString
}; # readRow


proc mcdu::ann {type state} {
  # Controls the annunciators. Not all annunciators are supported on all
  # hardware, but this function should never cause a failure.
  # <in> type = One of "dspy", "fail", "msg", "ofst", "exec", or "all".
  # <in> 0 for off, 1 for on.
  variable _ANNUNCIATORS
  variable _ANNUNCIATORS_MAGENTA
  global IMGDIR

  # State sanity check.
  if {$state!="1"} {set state 0}

  # Special case for all annunciators. Take care that the annunciator
  # switching might be implemented by key strokes; take time in between.
  if {$type=="all"} {
    after 50 "mcdu::ann dspy $state"
    after 100 "mcdu::ann exec $state"
    after 150 "mcdu::ann fail $state"
    after 200 "mcdu::ann msg $state"
    after 250 "mcdu::ann ofst $state"
    # Set up a non-busy-waiting delay of 300 ms to keep the event loop alive.
    after 300 {set mcdu::_ANNUNCIATORS 0}
    vwait mcdu::_ANNUNCIATORS
    return
  }

  # Annunciators on the bitmap. The bitmaps have already been loaded at init
  # time, but their availability may not be assumed.
  # Bitmap available?
  if {[lsearch [image names] "ann_$type"]>-1} {
    # Yes. Take it off or put it up?
    if {$state==0} {
      # Take it off.
      .c delete ann_$type
    } else {
      # Put it up. Fetch the display mode settings.
      array set mode [readMode [prefs::get display.mode]]
      set x [lindex $mode(ann_$type) 1]
      set y [lindex $mode(ann_$type) 2]
      .c create image $x $y -anchor nw -image ann_$type -tag ann_$type
      # Exception: raise the KEY_exec bitmap on top again, so that it remains
      # hot clickable.
      .c raise KEY_exec
    }
  }

  # Annunciators on the Meta Block.
  if {$state==0} {
    # Take it off.
    .c itemconfigure "ann_meta_$type" -fill [getCellColour -1 -1]
  } else {
    # Put it up.
    switch $type {
      "fail"  {set colour "red"}
      default {set colour "white"}
    }
    .c itemconfigure "ann_meta_$type" -fill $colour
  }
  
  # Annunciators on the iconified MCDU. Only when floating icon is present.
  if {[winfo exists .icon.f.l]} {
    if {$state==0} {
      .icon.f.l configure -foreground green -background black
    } else {
      switch $type {
        "msg"  {.icon.f.l configure -foreground black -background white}
        "fail" {.icon.f.l configure -foreground black -background red}
      }
    }
  }

  # Different actions depending on hardware setting.
  set keyboard [prefs::get keyboard]
  if {$keyboard==""} {set keyboard PS1}
  switch $keyboard {
    MAGENTA {
      # FAIL, MSG, and EXEC annunciators via the keyboard LEDs.
      # The FAIL annunciator doubles as OFST for the moment.
      switch $type {
        # Windows-only for now. Who else would need it ...?
	exec {win::keybd capslock $state}
	fail {win::keybd numlock $state}
	msg  {win::keybd scrolllock $state}
	ofst {win::keybd numlock $state}
      }
      after 100 {set mcdu::_ANNUNCIATORS_MAGENTA 0}
      vwait mcdu::_ANNUNCIATORS_MAGENTA
    }
  }; # switch keyboard

  # Engravity annunciators.
  engr::ann $type $state
}; # ann


proc mcdu::getRow {row} {
  # Return the character content of the given row.
  # <in> row = 0..13.
  # <out> = A 24-character string with the verbatim row content.
  # Note: this procedure will not correctly handle MCDU displays that are
  # different from 14 rows of 24 characters, but in PS1 FMC mode, this will
  # not happen.
  set rowString ""
  for {set col 0} {$col<24} {incr col} {
    append rowString [.c itemcget "$row,$col" -text]
  }
  return $rowString
}; # getRow


proc mcdu::getLabel {row side} {
  # Compute and return the label on (row,side). A label is the string from
  # the edge of the display to the first two spaces, exclusive, and trimmed.
  # <in> row = 0..13.
  # <in> side = l, r.
  # <out> = A plain, trimmed string.
  # Start with the full display row string.
  set label [getRow $row]
  if {$side=="l"} {
    # Scan for first two spaces.
    set pos [string first "  " $label]
    if {$pos>-1} {
      # Cut off at delimiter.
      set label [string range $label 0 $pos]
    }
  } else {
    # Scan for last two spaces.
    set pos [string last "  " $label]
    if {$pos>-1} {
      # Cut off at delimiter.
      set label [string range $label $pos end]
    }
  }
  return [string trim $label]
}; # getLabel


##### MCDU ACTIONS ########################################################


proc mcdu::mode {{mode ""}} {
  # Initiates a mode switch. The current mode is tacked onto most events and
  # appears in the GUI title bar/process icon.
  # All internal modes start with "#".
  # <in> Optionally, the new mode. If left out, only the current mode is
  #      returned.
  # <out> The new mode after the statement has been executed.
  variable _MODE

  if {$mode!=""} {
    # Switch the mode.
    set _MODE $mode
  }

  # Some modes do not need to be posted in the title bar.
  if {$mode==""} {
  } elseif {[string range $mode 0 0]=="#"} {
    wm title . "MCDU"
  } elseif {$mode=="reset"} {
    wm title . "MCDU"
  } else {
    set mode [string toupper $mode]
    wm title . "$mode - MCDU"
  }

  return $_MODE
}; # mode


proc mcdu::reset {} {
  # Initialises the whole MCDU, from a logical point of view (not the GUI).
  variable MODULES
  variable FUNCTIONKEYS

  # Send a keystroke with the RESET mode. The keystroke itself is a dummy.
  mode reset
  sendKey lsk1l

  set MODULES {}
  catch {unset FUNCTIONKEYS}
  array set FUNCTIONKEYS {}
  clear

  # Fetch the display mode settings.
  array set mode [readMode [prefs::get display.mode]]

  if {$mode(splash.skip)} {
    # With no splash screens, delay the startup a bit to give subsystems the
    # chance to register.
    sleep 3
  } else {
    # The MCDU can accept registration requests while displaying the test
    # patterns. This leads to a nice full menu, straight after reset.
    after 100 {mcdu::test init pattern}
    after 2000 {mcdu::test init about}
    sleep 4
  }

  # See whether it makes sense to skip the initial MENU page.
  if {$mode(menu.skip)} {
    # Is there currently any subsystem registered?
    if {[llength $MODULES]>0} {
      # Yes, go directly to the first module, skip MENU.
      mode [lindex $MODULES 0]
      after 500 {mcdu::sendKey menu}
      return
    }
  }

  # No menu.skip or no module available, show the MENU page.
  after 500 {mcdu::menuPage}
}; # reset


proc mcdu::shutdown {} {
  # Attempts to shut down the complete computer. This is not guaranteed to
  # work, especially not under Win98 which requires much more than just a
  # simple call.
  message add "shutting down..."
  update
  update idletasks
  after 2000
  win::shutdown
}; # shutdown


proc mcdu::powerEvent {stdbyPower} {
  # Reacts to incoming (sim) power events.
  # Probably we should connect to a bus or two instead of to the Stdby Power
  # Switch, but at least this is in line with the GoFlight panels.
  # <in> stdbyPower = 1=off, 2=auto, 3=bat.
  variable POWER

  # The MCDU is not really shut off or even reset when the power drops.
  # It is just a screen blank.
  if {$POWER==-1} return
  if {$POWER && $stdbyPower>1} return
  if {!$POWER && $stdbyPower==1} return

  if {$POWER && $stdbyPower==1} {
    # Power was just turned off. Blank the screen as if the CRT quits.
    set origBrightness [prefs::get display.brightness]
    message empty ""
    scratch "   "
    ann all 0
    clear
    prefs::put display.brightness 10
    setDisplayBrightness
    display 6 11 "."
    update
    after 1000
    for {set i 10} {$i>=0} {incr i -1} {
      prefs::put display.brightness $i
      setDisplayBrightness
      after 50
    }
    clear
    prefs::put display.brightness $origBrightness
    set POWER 0
    return
  }
  
  if {!$POWER && $stdbyPower>1} {
    # Switch the CRT on.
    after 1000
    set origBrightness [prefs::get display.brightness]
    prefs::put display.brightness 0
    setDisplayBrightness
    menuPage
    after 1000
    for {set i 0} {$i<$origBrightness} {incr i} {
      prefs::put display.brightness $i
      setDisplayBrightness
      after 200
    }
    prefs::put display.brightness $origBrightness
    setDisplayBrightness
    set POWER 1
  }
}; # powerEvent


proc mcdu::clear {{title ""}} {
  # Clears the display. Does not touch the scratch pad. Resets the colour to
  # the default.
  # <in> title = Optionally, the page title, centered.
  variable ROWS
  variable COLS

  set emptyLine [string repeat " " $COLS]
  colour
  for {set row 0} {$row<[expr $ROWS-1]} {incr row} {
    display $row 0 $emptyLine
  }
  engr::clear
  display 0 c $title
  update
}; # clear


proc mcdu::scratch {content} {
  # Put the given content in the scratch pad. If there is a message
  # in the queue, buffer the scratch pad but postpone display until the
  # message queue is empty. Colour will always be the mode default.
  # <in> content = The new scratch pad content, typically max. 24 characters.
  #                It will always be converted to lower case.
  variable SCRATCHPAD
  variable MESSAGES
  variable ROWS
  variable COLS

  set maxrow [expr $ROWS-1]
  set maxcol [expr $COLS-1]
  set emptyLine [string repeat " " $COLS]
  set content [string tolower [string range "$content" 0 $maxcol]]
  set SCRATCHPAD $content

  if {$MESSAGES=={}} {
    colour
    display $maxrow 0 [string tolower [string range "$content$emptyLine" \
                       0 $maxcol]]
  }
}; # scratch


proc mcdu::message {cmd {msg ""}} {
  # Add or remove a message from the queue. New messages are always added
  # to the end; there is no priority mechanism (yet).
  # <in> cmd = "empty", "add" or "remove".
  # <in> msg = Optionally, the message. If blank and the command is
  #            "remove", the first message (the one currently on display)
  #            is deleted. A blank "add" is ignored. The msg is always
  #            converted to lower case.
  variable MESSAGES
  variable SCRATCHPAD
  variable ROWS

  set msg [string tolower $msg]

  switch -- $cmd {
    empty {
      set MESSAGES {}
    }
    add {
      if {$msg!=""} {lappend MESSAGES $msg}
    }
    remove {
      if {$msg==""} {set msg [lindex $MESSAGES 0]}
      set newQueue {}
      foreach m $MESSAGES {
        if {$m!=$msg} {
          lappend newQueue $m
        }
      }
      set MESSAGES $newQueue
    }
  }

  if {$MESSAGES==""} {
    # Uncover the scratch pad and extinguish the annunciator
    # if the last message got deleted.
    display [expr $ROWS-1] 0 "$SCRATCHPAD                                     "
    ann msg 0
  } else {
    # Display the first message of the queue and light the annunciator.
    display [expr $ROWS-1] 0 "[lindex $MESSAGES 0]                            "
    ann msg 1
    # Bell only if not DELETE.
    if {[lindex $MESSAGES 0]!="delete"} bell
  }
}; # message


proc mcdu::module {cmd modName} {
  # Add or remove a module to/from the module list. A duplicate module
  # or unknown module is not added again, but the acknowledgement is
  # given. This enables modules to re-incarnate after their death.
  variable MODULES
  variable FUNCTIONKEYS
  variable PASSTHROUGH

  switch -- $cmd {

    add {
      # Duplicate entry?
      set found 0
      foreach m $MODULES {
        if {$m==$modName} {
          set found 1
          break
        }
      }
      if {!$found} {
        lappend MODULES $modName
        set PASSTHROUGH($modName) 0
      }
      ps1::toBroker "put mcdu.event \{modadded $modName\}"
      if {[mode]=="#select"} {
        # Repaint the menu.
        menuPage
      }
      # Just in case the module took a hike for a moment.
      message remove "$modName unavailable"
    }

    remove {
      set newModules {}
      set found 0
      foreach m $MODULES {
        if {$m!=$modName} {
          lappend newModules $m
        } else {
          set found 1
        }
      }
      set MODULES $newModules
      if {$found} {
        ps1::toBroker "put mcdu.event \{modremoved $modName\}"
        # Deregister function keys that might be linked to this mode.
        foreach key [array names FUNCTIONKEYS] {
          if {$FUNCTIONKEYS($key)==$modName} {
            unset FUNCTIONKEYS($key)
          }
        }
        # Remove the PASSTHROUGH indicator.
        unset PASSTHROUGH($modName)
        if {[mode]=="#select"} {
          # Repaint the menu.
          menuPage
        }
        if {[mode]==$modName} {
          # The current mode got removed. Show main menu and complain.
          after 500 "mcdu::message add \"$modName unavailable\""
          menuPage
        }
      }
    }

  }; # switch
}; # module


proc mcdu::moduleKey {mode keys} {
  # Registers function keys to a module, for direct access. One key can only
  # point to one module; later registrations overwrite older ones.
  # <in> mode = The module name to register the function key(s) for.
  # <in> keys = Tcl list of keysyms (initref, vnav, ...) to associate with
  #             the given module name.
  variable MODULES
  variable FUNCTIONKEYS

  # Only works for existing modules.
  if {[lsearch $MODULES $mode]==-1} return

  foreach key $keys {
    set FUNCTIONKEYS($key) $mode
  }
}; # moduleKey


proc mcdu::keyBoard {char code {mod none}} {
  # Offer a hook for keytable translation.
  # <in> char = %K keysym.
  # <in> code = %k keysym.
  # <in> mod = Optionally, modifier, "Control", "Shift" etc.
  variable FONT_SIZE
  variable XLAT

  # Control key held down?
  if {$mod!="none"} {
    set code "$mod-$code"
  }
  # When in Test Mode, post the keypress in the scratch pad.
  if {[mode]=="#test"} {
    scratch "$char/$code            "
  }

  # Only process known key strokes. Ignore all others.
  if {[info exists XLAT($code)]} {
    keyPressed $XLAT($code)

    # Get the coordinates of the virtual key and flash a rectangle there.
    # This code is duplicated in mcdu::mouse.
    set keysym "KEY_$XLAT($code)"
    set keyCoords [.c coords $keysym]
    set kx [lindex $keyCoords 0]
    set ky [lindex $keyCoords 1]
    # The "radius" of the rectangle varies with the master font size.
    set r [expr $FONT_SIZE*0.6]
    .c create rectangle \
      [expr $kx-$r] [expr $ky-$r] [expr $kx+$r] [expr $ky+$r] \
      -outline white -tag CLICK_FLASH
    update idletasks
    after 100 {.c delete CLICK_FLASH}
  }
}; # keyBoard


proc mcdu::mouse {x y b} {
  # Offer a hook for mouse click interception.
  # <in> x = The X position of the mouse cursor at click time.
  # <in> y = The Y position of the mouse cursor at click time.
  # <in> b = The number of the button that was pressed.
  variable DRAG
  variable FONT_SIZE

  if {$b==3} {
    # Remember where MB3 was first clicked, for possible subsequent dragging.
    set DRAG [list $x $y]
    set x [format "%04d" $x]
    set y [format "%04d" $y]
    # MB3 mouse clicks change the cursor to the "move" cursor. Notice that
    # this cursor won't change back into normal when the button is released;
    # it only does so when the cursor is moved. Nearly nobody will notice this.
    . configure -cursor fleur
    # Repeat to avoid mouseHide removing the cursor again after 150 ms.
    after 200 {. configure -cursor fleur}
    return
  }

  # MB1 clicks can be either on a virtual key or just display/decoration area.
  # Use FONT_SIZE (which is always negative, i.e. pixels and not pica
  # points) to create the halo size for "find closest".
  set id [.c find closest $x $y [expr -$FONT_SIZE*0.6] 1]
  set keysym [lindex [.c gettag $id] 0]; # to avoid getting "current".
  if {[string range $keysym 0 3]=="KEY_"} {
    # This was a click on a virtual key. Show some user feedback:
    # get the coordinates of the virtual key and flash a rectangle there.
    # This code is duplicated in mcdu::keyBoard.
    set keyCoords [.c coords $keysym]
    set kx [lindex $keyCoords 0]
    set ky [lindex $keyCoords 1]
    # The "radius" of the rectangle varies with the master font size.
    set r [expr $FONT_SIZE*0.6]
    .c create rectangle \
      [expr $kx-$r] [expr $ky-$r] [expr $kx+$r] [expr $ky+$r] \
      -outline white -tag CLICK_FLASH
    update idletasks
    after 100 {.c delete CLICK_FLASH}
    # Look up the associated key symbol for the virtual key press.
    set keysym [lindex [string range $keysym 4 end] 0]
    if {$keysym=="kbd"} {
      # kbd never is a true key press, but needs to be handled by the
      # GUI only.
      virtualKeyboard
    } else {
      # Emulate a key press on the keyboard.
      keyPressed $keysym
    }
  } else {
    # It wasn't recognised as a key press. If in test mode, show the
    # click coordinates on the scratch pad.
    if {[mode]=="#test"} {
      scratch "x/$x y/$y"
    }
  }; # if a keypress
}; # mouse


proc mcdu::mouseIcon {x y {double ""}} {
  # Offer a hook for mouse click interception.
  # <in> x = The X position of the mouse cursor at click time.
  # <in> y = The Y position of the mouse cursor at click time.
  variable DRAGICON
  variable DRAGNOTYET
  variable ICONIC
  set DRAGNOTYET [clock clicks -milliseconds]

  # Remember where mouse clicked, for possible subsequent dragging.
  set DRAGICON [list $x $y]

  # Always raise the main MCDU window if it is currently mapped. This avoids
  # needless four-click actions to get a covered MCDU back up.
  raise .

  if {$double=="double"} {
    # Double-click detected. Either withdraw the main MCDU from the screen
    # or make it appear again, on top of the Z-order.
    if {$ICONIC} {
      global tcl_platform
      if {$tcl_platform(platform)=="unix"} {
        # On Linux (at least Ubuntu 10.10), we need to force override redirect
        # for the main toplevel (MCDU) window here, else it will get higher
        # on-top priority as the icon. Weird!
        # Fetch the display mode settings.
        array set mode [readMode [prefs::get display.mode]]
        wm overrideredirect . $mode(titlebar.hide)
        wm state . normal
        wm overrideredirect . 0
      } else {
        wm state . normal
      }
      set ICONIC 0
    } else {
      wm state . withdrawn
      set ICONIC 1
    }
  }
}; # mouseIcon


proc mcdu::motion {} {
  # Make the mouse cursor visible (one finger pointer hand). This is
  # interesting when a hidden cursor suddenly gets picked up by a mouse
  # instead of a finger.
  . configure -cursor {}
}; # motion


proc mcdu::drag {x y} {
  # Offer a hook for mouse drag interception.
  # <in> x = The X position of the mouse cursor during dragging.
  # <in> y = The Y position of the mouse cursor during dragging.
  variable DRAG

  # Get relative displacement from the first button-3-down position.
  set dx [expr $x-[lindex $DRAG 0]]
  set dy [expr $y-[lindex $DRAG 1]]
  # Fetch the current position of the (toplevel) window incl. decorations.
  set oldgeom [split [wm geometry .] "+"]
  set wx [lindex $oldgeom 1]
  set wy [lindex $oldgeom 2]
  # Reposition according to the drag.
  set newgeom "+[expr $wx+$dx]+[expr $wy+$dy]"
  wm geometry . $newgeom
}; # drag


proc mcdu::dragIcon {x y} {
  # Offer a hook for mouse drag interception.
  # <in> x = The X position of the mouse cursor during dragging.
  # <in> y = The Y position of the mouse cursor during dragging.
  variable DRAGICON
  variable DRAGNOTYET

  # Delay 150 ms to avoid moving the icon when a finger tap isn't staying on
  # the same spot. This does not happen easily with a mouse, but is very
  # common for a finger on a touch screen.
  if {$DRAGNOTYET>[expr [clock clicks -milliseconds]-150]} return

  # Get relative displacement from the first button-1-down position.
  set dx [expr $x-[lindex $DRAGICON 0]]
  set dy [expr $y-[lindex $DRAGICON 1]]
  # Fetch the current position of the (toplevel) window incl. decorations.
  set wx [winfo x .icon]
  set wy [winfo y .icon]
  # Reposition according to the drag.
  set newgeom "+[expr $wx+$dx]+[expr $wy+$dy]"
  wm geometry .icon $newgeom
}; # dragIcon


proc mcdu::mouseHide {} {
  # If the mouse.autohide option is selected in the mode configuration file,
  # suppresses the mouse cursor when over the MCDU window (only).
  # Is typically called after each mouse click and each keyboard keypress,
  # so that touch screens can be nicely supported.

  # Fetch the display mode settings.
  array set mode [readMode [prefs::get display.mode]]

  if {$mode(mouse.autohide)==1} {
    # Hide mouse pointer. "none" seems to be undocumented and only exists in
    # Tk 8.5 and higher.
    . configure -cursor none
  }
}; # mouseHide


proc mcdu::keyPressed {keysym} {
  # A key on the MCDU has been pressed. This procedure handles all of them.
  # Some keys are "hot" and never go outside this routine, such as
  # all of the scratch pad handling (unless in passthrough mode).
  # Others may or may not go out, depending on the MCDU mode.
  # <in> keysym = The MCDU keysym.
  global ETCDIR
  variable SCRATCHPAD
  variable MESSAGES
  variable MODULES
  variable LAYOUT
  variable FUNCTIONKEYS
  variable _KEYINT
  variable COLS
  variable PASSTHROUGH

  # Do not handle a new key press when the current key press is still
  # being handled.
  if {$_KEYINT} return
  set _KEYINT 1

  # Hide the mouse pointer when needed. A small delay is used to suppress
  # finger motion after tap on a touch screen.
  after 150 {mcdu::mouseHide}

  # Brightness adjustment keys are never processed as keys.
  if {[string range $keysym 0 2]=="brt"} {
    set brightness [prefs::get display.brightness]
    if {$brightness==""} {set brightness 7}
    if {$keysym=="brt-"} {
      incr brightness -1
      if {$brightness<0} {set brightness 0}
    } else {
      incr brightness
      if {$brightness>10} {set brightness 10}
    }
    prefs::put display.brightness $brightness
    setDisplayBrightness
    set _KEYINT 0
    return
  }

  if {[prefs::get mcdu.logging]!=""} {
    # Log the keystroke and a few relevant data items.
    set header ""
    set data   ""
    if {[string range $keysym 0 2]=="lsk"} {
      # For LSKs, get the header and data associated with the key.
      set lskNr [string index $keysym 3]
      set headerLine [expr ($lskNr*2)-1]
      set dataLine   [expr ($lskNr*2)]
      set lskSide [string index $keysym 4]
      set header [getLabel $headerLine $lskSide]
      set data   [getLabel $dataLine   $lskSide]
    }

    # Get the scratch pad content.
    set scratch [string trim [getRow 13]]

    # Write the log entry.
    set log [open "$ETCDIR/mcdu.log" a]
    puts $log "[clock milliseconds] $keysym {$header} {$data} {$scratch}"
    close $log
  }; # if MCDU logging enabled

  # The "menu" key is handled by the MCDU for main menu selection, except
  # when already in MENU mode.
  if {$keysym=="menu"} {
    # In MENU (#select) mode?
    if {[mode]=="#select"} {
      # FMC module present?
      foreach module $MODULES {
        if {$module=="fmc"} {
          # MENU key in MENU mode with FMC mode present = switch to FMC mode
	  # and send MENU key to PS1 FMC.
	  mode fmc
          fmc::keystroke menu
	  set _KEYINT 0
          return
        }; # FMC mode?
      }; # foreach mode
    }; # mode=#select?
    menuPage
    set _KEYINT 0
    return
  }

  # Modes can register function keys for them. They work in any MCDU mode
  # and may cause a mode switch.
  if {[info exists FUNCTIONKEYS($keysym)]} {
    # This function key was indeed registered.
    set newMode $FUNCTIONKEYS($keysym)
    if {$newMode=="fmc"} {
      # The FMC mode is internal; needs special shortcut.
      mode fmc
      fmc::keystroke $keysym
      set _KEYINT 0
      return
    }
    mode $newMode
    sendKey $keysym
    set _KEYINT 0
    return
  }

  # In case the current mode is SELECT (= MENU page displayed), just hang
  # around until the user selects a mode. Modes include the internal MCDU
  # configuration and test page, "mcdu".
  if {[mode]=="#select"} {
    # On MENU page.
    if { (($keysym>="lsk1l") && ($keysym<="lsk6l")) ||
         (($keysym>="lsk1r") && ($keysym<="lsk5r"))} {
      # Mode LSK pressed. Select this mode.
      set pos [string index $keysym 3]
      if {[string index $keysym 4]=="r"} {
        incr pos 5
      } else {
        incr pos -1
      }
      set mode $LAYOUT($pos)
      if {$mode!=""} {
        # Switch mode. Start with a clean display.
        clear
        mode $mode
	if {$mode=="fmc"} {
	  # FMC mode, reset the PS1 FMC communication.
	  fmc::reset
	} else {
	  # Other mode, init by sending the MENU event to the module.
          sendKey menu
	}
      }
    } elseif {$keysym=="lsk6r"} {
      # Internal MCDU config page selected.
      config menu
    } elseif {$keysym=="clr"} {
      set msg [lindex $MESSAGES 0]
      message remove $msg
      ps1::toBroker "put mcdu.cmd \{msgremove [mode] \{$msg\}\}"
    }
    set _KEYINT 0
    return
  }; # if on MENU page

  # If the current mode is a passthrough mode, handle the key directly.
  if {[info exists PASSTHROUGH([mode])]} {
    if {$PASSTHROUGH([mode])} {
      # Send the key straight to the subsystem...
      sendKey $keysym
      # ... and quit working on it.
      set _KEYINT 0
      return
    }
  }

  # If in the special FMC mode, pass on all keys directly to the PS1 FMC.
  # Note that MENU has already been handled above.
  if {[mode]=="fmc"} {
    fmc::keystroke $keysym
    set _KEYINT 0
    return
  }; # if in FMC mode

  # PgUp and PgDn keys are sent directly to the applications.
  if {$keysym=="prevpage"} {
    sendKey "prevpage"
    set _KEYINT 0
    return
  }
  if {$keysym=="nextpage"} {
    sendKey "nextpage"
    set _KEYINT 0
    return
  }

  # Same for line select keys.
  if {[string range $keysym 0 2]=="lsk"} {
    sendKey $keysym
    set _KEYINT 0
    return
  }

  # The DEL key actually creates a message "delete".
  if {$keysym=="del"} {
    # Second press of the DEL key?
    if {[lindex $MESSAGES 0]=="delete"} {
      # Cancel the DELETE annunciation.
      message remove
      set _KEYINT 0
      return
    }
    # First press of DEL and scratch pad empty?
    if {$SCRATCHPAD=="" && $MESSAGES==""} {
      message add "delete"
      set _KEYINT 0
      return
    }
  }

  # Rest goes to the scratch pad, either directly or as message
  # manipulation.

  if {$keysym=="clr"} {
    # Is there a message on display right now?
    if {$MESSAGES!=""} {
      set msg [lindex $MESSAGES 0]
      message remove $msg
      ps1::toBroker "put mcdu.cmd \{msgremove [mode] \{$msg\}\}"
      set _KEYINT 0
      return
    }
    # Cut one char off the right end of the scratchpad.
    set pad $SCRATCHPAD
    set len [string length $pad]
    if {$len>0} {
      set pad [string range $pad 0 end-1]
    }
    scratch $pad
    set _KEYINT 0
    return
  }

  # During message display, no characters are passed on to the scratch pad.
  if {$MESSAGES!=""} {
    set _KEYINT 0
    return
  }

  # The +/- key is an exception. If the last character of the scratch
  # pad currently is a minus, change it into a plus and vv. Don't append.
  if {$keysym=="+/-"} {
    if {[string index $SCRATCHPAD end]=="-"} {
      set SCRATCHPAD [string range $SCRATCHPAD 0 end-1]
      set keysym "+"
    } elseif {[string index $SCRATCHPAD end]=="+"} {
      set SCRATCHPAD [string range $SCRATCHPAD 0 end-1]
      set keysym "-"
    } else {
      set keysym "-"
    }
  }

  # The sp key needs to be translated, too.
  if {$keysym=="sp"} {
    set keysym " "
  }

  # Only uppercase allowed.
  set keysym [string toupper $keysym]

  # Anything weird left? Kill it.
  if {[string first $keysym "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-+/ "]==-1} {
    set _KEYINT 0
    return
  }

  # No leading spaces on the scratch pad.
  if {$SCRATCHPAD=="" && $keysym==" "} {
    set _KEYINT 0
    return
  }

  # Append the new character.
  if {[string length $SCRATCHPAD]<$COLS} {
    set pad "$SCRATCHPAD$keysym"
    scratch $pad
  }

  set _KEYINT 0
}; # keyPressed


proc mcdu::sendKey {key} {
  # Send a key press to the subscribed applications, which might be the
  # internal CONFIG or TEST mode. (SELECT is handled by keyPressed).
  # <in> key = The key press, in mnemonic form.
  variable SCRATCHPAD
  variable COUNTER
  variable MESSAGES

  # If there is a message on display, check whether it is "DELETE".
  # In that case, replace the actual scratch pad with "DELETE" in capitals
  # so that applications can react to the delete command.
  if {[lindex $MESSAGES 0]=="delete"} {
    set sp "DELETE"
  } else {
    set sp $SCRATCHPAD
  }

  if {[mode]=="#config"} {
    # The CONFIG mode stays internal.
    config $key $sp
  } elseif {[mode]=="#test"} {
    # The TEST mode stays internal.
    test $key $sp
  } elseif {[mode]=="#layout"} {
    # The MENU LAYOUT mode stays internal.
    menuLayout $key $sp
  } elseif {[string index [mode] 0]=="#"} {
    # All other internal modes don't accept keystrokes.
  } else {
    # Put the keystroke on the bus.
    set COUNTER [expr [incr COUNTER]%100]
    ps1::toBroker "put mcdu.event \{key [mode] $key \{$sp\} $COUNTER\}"
  }
}; # sendKey


proc mcdu::cmd {params} {
  # Handles incoming MCDU commands. This procedure is called with every
  # value change of the mcdu.cmd Broker key.
  # <in> params = The full Broker value of the mcdu.cmd key.
  variable PASSTHROUGH

  set cmd [lindex $params 0]
  set mode [lindex $params 1]
  set params [lrange $params 2 end]
  
  switch -- $cmd {

    ann {
      if {$mode!=[mode]} return
      ann [lindex $params 0] [lindex $params 1]
    }

    clear {
      if {$mode!=[mode]} return
      clear [lindex $params 0]
    }

    colour {
      if {$mode!=[mode]} return
      colour [lindex $params 0]
    }

    display {
      if {$mode!=[mode]} return
      display [lindex $params 0] [lindex $params 1] \
              [lindex $params 2] [lindex $params 3]
    }

    displayc {
      if {$mode!=[mode]} return
      displayc [lindex $params 0] [lindex $params 1] \
               [lindex $params 2] [lindex $params 3] [lindex $params 4]
    }

    modadd {
      module add $mode
    }

    modremove {
      module remove $mode
    }

    modkey {
      moduleKey $mode [lindex $params 0]
    }

    msgadd {
      # Check for global message.
      if {$mode!="all" && $mode!=[mode]} return
      message add [lindex $params 0]
    }

    msgremove {
      # Check for global message.
      if {$mode!="all" && $mode!=[mode]} return
      message remove [lindex $params 0]
    }

    passthrough {
      # Set the addressed mode to passthrough state. This needs to work also
      # when the mode is not yet selected active.
      # TODO we could check whether the addressed mode actually exists...
      set PASSTHROUGH($mode) 1
    }

    scratch {
      if {$mode!=[mode]} return
      scratch [lindex $params 0]
    }

    keyboard {
      # Keyboard mode change. Relies on further processors for correctness
      # check.
      keyboard $mode
    }

    key {
      # Externally generated key press, usually from MCDU hardware.
      # $mode here means mcdu_id!
      if {$mode!="" && $mode==[prefs::get mcdu.id]} {
        # This MCDU was sent a key.
        keyPressed [lindex $params 0]
      }
    }

    default {
      ps1::log "Unknown cmd ($cmd)"
    }
    
  }; # switch
}; # cmd


##### INTERNAL MCDU PAGES #################################################

proc mcdu::menuPage {} {
  # Display the MCDU main menu page, which contains all registered
  # modules. If there is a preferred layout, it is honoured as far as
  # possible. A maximum of eleven modules can be displayed.
  # The actual module selection via the LSKs happens in mcdu::keypressed.
  variable MODULES
  variable LAYOUT

  # Extinghuish all annunciators, clear the scratch pad, and empty the
  # message buffer.
  ann all 0
  scratch ""
  message empty

  # Display the menu page.
  array set LAYOUT [prefs::get menu.layout]
  # We have a list of preferred positions LAYOUT(pos)="mod" and a list of
  # available modules MODULES={mod1 mod2 mod3}.
  # First we distribute the available modules over the LAYOUT array,
  # filling up gaps if an unknown module is encountered.
  foreach module [lsort $MODULES] {
    # Search for this module in the layout array.
    set notFound 1
    foreach pos {0 1 2 3 4 5 6 7 8 9 10} {
      if {$LAYOUT($pos)==$module} {
        set notFound 0
        break
      }
    }
    if {$notFound} {
      # Put the module in the first available empty spot, or throw away.
      foreach pos {0 1 2 3 4 5 6 7 8 9 10} {
        if {$LAYOUT($pos)==""} {
          set LAYOUT($pos) $module
          break
        }
      }
    }
  }; # for each module available

  # Then we remove modules from LAYOUT that are not present in MODULES.
  foreach pos {0 1 2 3 4 5 6 7 8 9 10} {
    if {[lsearch -exact $MODULES $LAYOUT($pos)]==-1} {
      # Not found. Throw out.
      set LAYOUT($pos) ""
    }
  }
    
  # Lastly we display and select directly from LAYOUT.
  clear "menu"
  foreach pos {0 1 2 3 4 5 6 7 8 9 10} {
    set module $LAYOUT($pos)
    if {$module==""} continue

    # Display not more than eight characters per module name.
    set module [string range $module 0 7]

    if {$pos<6} {
      # Module in left column.
      display [expr $pos*2+2] 0 "<$module"
      if {$module==[mode]} {
        display [expr $pos*2+2] 10 "<act"
      }
    } else {
      # Module in right column.
      display [expr (($pos-6)*2)+2] r "$module>"
      if {$module==[mode]} {
        display [expr (($pos-6)*2)+2] 10 "act>"
      }
    }
  }

  # MCDU internal LSK: CONFIG>. Visible?
  array set mode [readMode [prefs::get display.mode]]
  if {$mode(config.prompt)!="none"} {
    display 12 r "config>" small
  }

  mode #select
}; # menuPage


proc mcdu::test {{char "init"} {sp ""}} {
  # Show a few test patterns on the display. Just for visual fun. The
  # NextPage/PrevPage keys cycle through the various options.
  # <in> char = Optionally, the pressed key. When "init", uses the
  #             scratch pad content for pattern selection.
  # <in> sp   = Optionally, the pattern to be shown.
  global VERSION
  variable _TEST

  scratch ""
  clear
  mode #test

  switch -- $char {
    init {
      # Test Mode has been called up.
      set _TEST $sp
    }
    nextpage {
      switch $_TEST {
	about   {set _TEST pattern}
	pattern {set _TEST demo}
	demo    {set _TEST atm}
	atm     {set _TEST display}
	display {set _TEST colour}
	colour {set _TEST about}
      }
    }
    prevpage {
      switch $_TEST {
        about   {set _TEST colour}
	pattern {set _TEST about}
	demo    {set _TEST pattern}
	atm     {set _TEST demo}
	display {set _TEST atm}
	colour {set _TEST display}
      }
    }
  }; # char switch

  testPattern $_TEST

}; # test


proc mcdu::testPattern {pattern} {
  # Show a test pattern on the display.
  # <in> pattern = The pattern to display, one of "about", "pattern",
  #                "demo", "atm", "display", or "colour".
  global VERSION
  variable ROWS
  variable COLS
   
  switch -- $pattern {

    about {
      # Basic program information.
      message remove all
      clear "747-400 MCDU"
      display  3 c "V$VERSION   (c) 2010"
      display  5 c "Jeroen Hoppenbrouwers"
      display 11 c "www.hoppie.nl"
      scratch ""
    }

    pattern {
      # Character patterns and annunciators.
      set chars "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-+/%()!'#*^?:;~ ";
      while {[string length $chars]<[expr $ROWS*$COLS]} {
        append chars $chars
      }
      for {set row 0} {$row<$ROWS} {incr row} {
        set p [string range $chars [expr $row*$COLS] [expr (($row+1)*$COLS)-1]]
        display $row 0 $p
      }
      after 100 {mcdu::ann dspy 1}
      after 200 {mcdu::ann exec 1}
      after 300 {mcdu::ann fail 1}
      after 400 {mcdu::ann msg 1}
      after 500 {mcdu::ann ofst 1}
      after 1000 {mcdu::ann dspy 0}
      after 1100 {mcdu::ann exec 0}
      after 1200 {mcdu::ann fail 0}
      after 1300 {mcdu::ann msg 0}
      after 1400 {mcdu::ann ofst 0}
    }

    demo {
      # Typical FMC page.
      scratch ""
      clear
      display  0  4 "act econ crz"
      display  0 -4 "2/3" small
      display  1  0 " crz alt         step to"
      display  2  0 "fl330"
      display  2  r "fl370" small
      display  3  0 " econ spd             at"
      display  4  0 ".810        1004z/  71"
      display  4 16 "z" small
      display  4  r "nm" small
      display  5  0 " epr       ksea eta/fuel"
      display  6  0 "1.35         2003z/ 50.6"
      display  6 17 "z" small
      display  7  0 " step size    opt    max"
      display  8  0 "icao         fl362 fl384"
      display  9  0 "------------------------" large
      display 10  r "eng out>"
      display 12  r "lrc>"
    }

    atm {
      # ATM mode for fun.
      display  0 c "cash withdrawal"
      display  2 r "USD  10>"
      display  3 0 " PIN"
      display  4 0 "!!!!"
      display  7 0 " receipt"
      display  8 0 "<yes"
      display  8 5 "select" small
      display  4 r "USD  20>"
      display  6 r "USD  50>"
      display  8 r "USD 100>"
      display 10 r "USD 200>"
      display 11 0 " transaction"
      display 12 0 "<abort          USD 500>"
    }
    
    display {
      # Display elements.
      clear "page title"
      display  0  r "n/t" small
      display  1  0 " line title 1 line title"
      display  2  0 "line data   1  line data"
      display  3  0 " line title 2 line title"
      display  4  0 "line data   2  line data"
      display  5  0 " line title 3 line title"
      display  6  0 "line data   3  line data"
      display  7  0 " dashes     4 box prompt"
      display  8  0 "---/----    4     !!!!!!"
      display  9  0 " line title 5 line title"
      display 10  0 "large font  5"
      display 10  r "small font" small
      display 11  0 "------------------------"
      display 12  0 "<pg prompt    pg prompt>"
      scratch "scratch pad line"
    }

    colour {
      # Colour modes.
      clear "arinc 739 colour modes"
      displayc  1 0 default "black"
      displayc  2 0 black   "black text"
      displayc  3 0 default "cyan"
      displayc  4 0 cyan    "cyan text"
      displayc  5 0 default "red"
      displayc  6 0 red     "red text"
      displayc  7 0 default "yellow"
      displayc  8 0 yellow  "yellow text"
      displayc  9 0 default "green"
      displayc 10 0 green   "green text"
      displayc  1 r default "magenta"
      displayc  2 r magenta "magenta text"
      displayc  3 r default "amber"
      displayc  4 r amber   "amber text"
      displayc  5 r default "white"
      displayc  6 r white   "white text"
      scratch "scratch pad line"
    }
  }; # switch

}; # testPattern


proc mcdu::config {{char ""} {sp ""}} {
  # Show internal MCDU config page or handles "config" mode key strokes.
  # <in> char = Optionally, the pressed key. When "menu", assumes
  #             initial mode selection (displays MCDU CONFIG page).
  # <in> sp   = Optionally, the current scratch pad content.
  variable RESET_ARMED
  variable QUIT_ARMED
  variable SHUTDOWN_ARMED
  variable PASSWORD_OK
  variable DISPLAYWIDTH
  variable KEYBOARD
  variable DISPLAY_MODE

  set broker [prefs::get broker.host]
  if {$broker==""} {set broker "off"}

  set keyboard [prefs::get keyboard]
  if {$keyboard==""} {set keyboard "PS1"}
  if {$KEYBOARD==""} {set KEYBOARD $keyboard}

  set displayMode [prefs::get display.mode]
  if {$displayMode==""} {set displayMode [config_nextMode ""]}
  if {$DISPLAY_MODE==""} {set DISPLAY_MODE $displayMode}

  set engravity [prefs::get engravity.port]

  # Is the CONFIG page password-protected? If so, then only harmless
  # config features are available, and 1R shows a PASSWORD prompt.
  array set mode [readMode [prefs::get display.mode]]
  set password $mode(config.password)

  switch -- $char {

    "menu" {
      # Init CONFIG mode.
      mode #config
      set RESET_ARMED 0
      set QUIT_ARMED 0
      set SHUTDOWN_ARMED 0
      set PASSWORD_OK 0
      set KEYBOARD $keyboard
      after 10 {mcdu::config}
    }
    "" {
      # Regular processing.
      clear "mcdu configuration"
      # Harmless config options.
      display  1 0 " mcdu"
      display  2 0 "<reset"
      display 11 0 " show"
      display 12 0 "<test pages"

      display  9 r "mcdu prgrm"
      display 10 r "quit>"
      display 11 r "computer"
      display 12 r "shutdown>"
      if {$PASSWORD_OK || $password==""} {
        # Harmful config options.
        display  3 0 " keyboard"
        display  4 0 "<next"
        display  5 0 " $KEYBOARD    "
        display  6 0 "<select"
        display  7 0 " engravity"
        display  8 0 $engravity
        display  9 0 " mcdu menu"
        display 10 0 "<layout"
        if {$ps1::BROKER=="0"} {
          display  1 r "broker ip"
        } else {
          display 1 r "conn to"
        }
        display  2 r $broker
        display  3 r "dspy mode"
        display  4 r "next>"
        display  5 r $DISPLAY_MODE
        display  6 r "select>"
        display  7 r "mcdu id"
        set id [prefs::get mcdu.id]
        if {$id==""} {set id "------"}
        display  8 r $id
      } else {
        # Password prompt.
        display  1 r "password"
        display  2 r "--------"
      }
    }

    lsk1l {
      # Reset.
      if {$RESET_ARMED} {
        reset        
      } else {
        display 2 0 "<confirm"
        set RESET_ARMED 1
      }
    }

    lsk2l {
      # Keyboard preselection.
      if {!$PASSWORD_OK && $password!=""} return
      set kbs {PS1 magenta linux off PS1}
      set kb [lsearch $kbs $KEYBOARD]
      incr kb
      set KEYBOARD [lindex $kbs $kb]
      display 5 0 " $KEYBOARD    "
    }

    lsk3l {
      # Keyboard activation.
      if {!$PASSWORD_OK && $password!=""} return
      prefs::put keyboard $KEYBOARD
      keyboard $KEYBOARD
    }

    lsk4l {
      # Up/downselect Engravity MCDU COM port.
      if {!$PASSWORD_OK && $password!=""} return
      if {$sp==""} {
        # Downselect current COM port.
        scratch $engravity
      } else {
        # Upselect. If "DELETE", replace by "off". Replace integers by COMn.
        switch -- $sp {
          "DELETE" {
            set sp "off"
            message remove
          }
          "1" {set sp "com1"}
          "2" {set sp "com2"}
          "3" {set sp "com3"}
          "4" {set sp "com4"}
        }
        # Check for valid string ("off" or "comN").
        if {![regexp {off|com[1-4]} $sp]} {
          message add "invalid entry"
          return
        }
        # Store and display the entry and kick the Engravity module.
        scratch ""
        prefs::put engravity.port $sp
        display 8 0 "$sp    "
        engr::init
      }
    }
    
    lsk5l {
      # Menu Layout page.
      if {!$PASSWORD_OK && $password!=""} return
      menuLayout
    }

    lsk6l {test init pattern}

    lsk1r {
      # Either enter a password, or up/downselect Broker host.
      if {!$PASSWORD_OK && $password!="" && $sp!=""} {
        # Password given. Okay?
        if {$password==$sp} {
          # Password OK. Register and repaint the CONFIG page.
          set PASSWORD_OK 1
          scratch ""
          after 10 {mcdu::config}
        } else {
          # Password not OK. Disregard.
          scratch ""
          message add "invalid entry"
        }
      } else {
        # Up/downselect Broker host.
        if {!$PASSWORD_OK && $password!=""} return
        if {$sp==""} {
          # Downselect current IP.
          scratch $broker
        } else {
          # Upselect. If "l", replace by "localhost". If "o", by "off".
          # If "DELETE", replace by "off".
          switch -- $sp {
            "l" {set sp "localhost"}
            "o" {set sp "off"}
            "DELETE" {
              set sp "off"
              message remove
            }
          }
          display 2 r "               "
          display 2 r $sp
          prefs::put broker.host $sp
          scratch ""
          if {$sp=="off"} {ps1::done}
          after 100 ps1::checkBrokerConnection
          after 200 mcdu::config
        }
      }
    }

    lsk2r {
      # Display mode preselection.
      if {!$PASSWORD_OK && $password!=""} return
      set DISPLAY_MODE [config_nextMode $DISPLAY_MODE]
      display 5 r "      $DISPLAY_MODE"
    }

    lsk3r {
      # Display mode activation.
      if {!$PASSWORD_OK && $password!=""} return
      prefs::put display.mode $DISPLAY_MODE
      init
      config
    }

    lsk4r {
      # Up/downselect this MCDU's identifier.
      if {!$PASSWORD_OK && $password!=""} return
      # An empty ID "" is displayed by "------" but not stored as such.
      set id [prefs::get mcdu.id]
      if {$sp=="" && $id!="------"} {
	# Downselect current ID.
        scratch $id
      } else {
	# Upselect. If "DELETE", replace by "".
	if {$sp=="DELETE"} {
          set sp ""
          message remove
	}
        prefs::put mcdu.id $sp
        display 8 r "               "
        if {$sp==""} {
          set sp "------"
        }
        display 8 r $sp
        scratch ""
      }

    }

    lsk5r {
      # Quit MCDU program.
      if {$QUIT_ARMED} {
        gui::done
      } else {
        display 10 r "  confirm>"
        set QUIT_ARMED 1
      }
    }

    lsk6r {
      # Shutdown.
      if {$SHUTDOWN_ARMED} {
        shutdown        
      } else {
        display 12 r "  confirm>"
        set SHUTDOWN_ARMED 1
      }
    }

  }; # switch

}; # config


proc mcdu::config_nextMode {currentMode} {
  # Returns the name of the next mode.
  # <in> currentMode = The current mode name, or "".
  # <out> Next mode name, or "" when no mode files available.
  global ETCDIR

  # Startup case: if no current mode is given, try "0800phot" as default.
  if {$currentMode==""} {
    if {[file exists "$ETCDIR/0800phot.cfg"]} {
      return "0800phot"
    }
  }

  # List all *.cfg files in the /etc directory.
  set modes [lsort [glob -nocomplain "$ETCDIR/*.cfg"]]
  # Make mode list "wrappable."
  set modes [concat $modes $modes]
  # Find the current mode.
  set cm [lsearch -glob $modes "*/$currentMode.cfg"]
  if {$cm==-1} {
    # Not found, take first available mode.
    set newMode [lindex $modes 0]
  } else {
    # Take the next available mode.
    set newMode [lindex $modes [expr $cm+1]]
  }

  # Strip directory path and extension off the mode name.
  return [file rootname [file tail $newMode]]
}; # config_nextMode


proc mcdu::menuLayout {{char ""} {sp ""}} {
  # Show internal MCDU menu layout page and handles the menu layout mode.
  # <in> char = Optionally, the pressed key. When not given, assumes
  #             initial mode selection.
  # <in> sp   = Optionally, the current scratch pad content.
  variable LAYOUT

  switch -- $char {

    "" {
      # Init MENU LAYOUT mode.
      unset LAYOUT
      array set LAYOUT [prefs::get menu.layout]
      mode #layout
      clear "menu layout config"
      foreach pos {0 1 2 3 4 5 6 7 8 9 10} {
        if {$pos<6} {
          display [expr ($pos*2)+2] L "<$LAYOUT($pos)"
        } else {
          display [expr (($pos-6)*2)+2] R "$LAYOUT($pos)>"
        }
      }; # for each menu position
      display 12 r "config>" small
    }

    lsk1l {menuLayout_select  0 $sp}
    lsk2l {menuLayout_select  1 $sp}
    lsk3l {menuLayout_select  2 $sp}
    lsk4l {menuLayout_select  3 $sp}
    lsk5l {menuLayout_select  4 $sp}
    lsk6l {menuLayout_select  5 $sp}
    lsk1r {menuLayout_select  6 $sp}
    lsk2r {menuLayout_select  7 $sp}
    lsk3r {menuLayout_select  8 $sp}
    lsk4r {menuLayout_select  9 $sp}
    lsk5r {menuLayout_select 10 $sp}

    lsk6r config

  }; # switch char

}; # menuLayout


proc mcdu::menuLayout_select {pos sp} {
  variable LAYOUT
  if {$sp==""} {
    # Downselect.
    scratch $LAYOUT($pos)
  } elseif {$sp=="DELETE"} {
    # Delete.
    set LAYOUT($pos) ""
    message remove "DELETE"
    prefs::put menu.layout [array get LAYOUT]
  } else {
    # Upselect.
    # If there is another module on the target, save it.
    set previous $LAYOUT($pos)
    set LAYOUT($pos) $sp
    scratch ""
    # Erase the other occurrence of this module, if any.
    # If the module replaced another one, swap them.
    foreach i {0 1 2 3 4 5 6 7 8 9 10} {
      if {$i!=$pos && $LAYOUT($i)==$sp} {
        set LAYOUT($i) $previous
      }
    }
    prefs::put menu.layout [array get LAYOUT]
  }
  after 200 mcdu::menuLayout
}

# EOF
