#!/bin/sh
# Copyright (C) 2011 Jeroen Hoppenbrouwers <hoppie@hoppie.nl>
# This program is free software; more information in the LICENSE file.

# \
exec wish8 "$0" "$@"

##### constants ###########################################################

set VERSION "2.32"
set APPNAME "MCDU"

# The application must be able to run both directly off the disk and from a
# wrapped (TclApp) virtual file system. The new virtual file system, unlike
# the older ProWrap system, cannot overlap a virtual and a real disk
# directory. This means that we cannot have user configuration files in a
# directory that is wrapped.
# It cannot be assumed that the CWD is predictable at startup time. So we
# calculate all paths in reference to the current script directory
# (directory of mcdu.tcl). Real disk paths use the same directory as
# mcdu.tcl, virtual file system paths include mcdu.tcl in the path.

# .../mcdu.tcl             Script file
# .../etc/                 User config files: cfg and ini
# .../img/                 User image files
# .../pkg/                 Third-party libraries that must be loaded
# .../mcdu.tcl/lib/        Wrapped library files (all types)

if {![namespace exists starkit]} {
  # Application runs unwrapped. Everything relative to mcdu.tcl.
  set LIBDIR "[file dirname [info script]]/lib"
  set ETCDIR "[file dirname [info script]]/etc"
  set IMGDIR "[file dirname [info script]]/img"
  set PKGDIR "[file dirname [info script]]/pkg"
} else {
  # Application runs wrapped. Some dirs are physically higher up the path.
  set LIBDIR "[file dirname [info script]]/lib"
  set ETCDIR "[file dirname [info script]]/../../../etc"
  set IMGDIR "[file dirname [info script]]/../../../img"
  set PKGDIR "[file dirname [info script]]/../../../pkg"
}

##### main ################################################################

# Load required libraries and configure them where required.
source "$LIBDIR/common/generic.tcl"

if {$tcl_platform(platform)=="windows"} {
  # Try to load the WinTcl Windows function library.
  if {[catch {load $LIBDIR/wintcl.dll} error]} {
    die $error
  }
  # Note: the custom wintcl.dll can be replaced by using twapi functions.
  # Example code:
  # twapi::GetKeyState 0x14
  # twapi::send_input { {key 0x14 0} {key 0x90 0} {key 0x91 0} }
  # twapi::shutdown_system
  # source "$PKGDIR/twapi-2.2.3.kit"
  # package require twapi
} else {
  # Set up simple dummy functions.
  namespace eval win {
    proc keybd {args} {}
    proc shutdown {args} {}
  }
}; # if platform is Windows

source "$LIBDIR/common/prefs.tcl"
prefs::init mcdu {
  broker.host 127.0.0.1
  display.brightness 7
  display.mode ""
  gui.position ""
  keyboard PS1
  engravity.port off
  menu.layout {0 "" 1 "" 2 "" 3 "" 4 "" 5 "" 6 "" 7 "" 8 "" 9 "" 10 ""}
  mcdu.id ""
  mcdu.logging ""
  icon.position ""
}
source "$LIBDIR/mcdu.tcl"
source "$LIBDIR/ps13.tcl"
source "$LIBDIR/gui.tcl"
source "$LIBDIR/fmc.tcl"
source "$LIBDIR/engr744.tcl"

# Enter the event loop.
